"""Tests of the functions in `sampling_tools.pxd` or their wrappers found in
`tofu.geom`.
"""
import numpy as np
import tofu as tf
import tofu.geom._GG as GG
from matplotlib.path import Path
from .testing_tools import compute_ves_norm
from .testing_tools import compute_min_max_r_and_z

# Global variables for testing
theta = np.linspace(0., 2. * np.pi, 100)
VPoly = np.array([2. + 1. * np.cos(theta), 0. + 1. * np.sin(theta)])
VIn = compute_ves_norm(VPoly)


# =============================================================================
# Vessel  - Line and VPoly sampling
# =============================================================================
def test01_ves_mesh_dlfroml():
    LMinMax = np.array([0., 10.])
    L, dLr, indL, N = GG.discretize_line1d(LMinMax, 20., DL=None,
                                           Lim=True, margin=1.e-9)

    assert np.allclose(L, [5.]) and dLr == 10. and np.allclose(indL, [0])
    assert N == 1
    L, dLr, indL, N = GG.discretize_line1d(LMinMax, 1., DL=None,
                                           Lim=True, margin=1.e-9)
    assert np.allclose(L, 0.5 + np.arange(0, 10)) and dLr == 1. and \
           np.allclose(indL, range(0, 10)) and N == 10
    DL = [2., 8.]
    L, dLr, indL, N = GG.discretize_line1d(LMinMax, 1., DL=DL,
                                           Lim=True, margin=1.e-9)
    assert np.allclose(L, 0.5 + np.arange(2, 8)) and dLr == 1. and \
           np.allclose(indL, range(2, 8)) and N == 10
    DL = [2., 12.]
    L, dLr, indL, N = GG.discretize_line1d(LMinMax, 1., DL=DL,
                                           Lim=True, margin=1.e-9)
    assert np.allclose(L, 0.5 + np.arange(2, 10)) and dLr == 1. and \
           np.allclose(indL, range(2, 10)) and N == 10
    DL = [2., 12.]
    L, dLr, indL, N = GG.discretize_line1d(LMinMax, 1., DL=DL,
                                           Lim=False, margin=1.e-9)
    assert np.allclose(L, 0.5 + np.arange(2, 12)) and dLr == 1. and \
           np.allclose(indL, range(2, 12)) and N == 10


def test02_discretize_vpoly(VPoly=VPoly):
    VIn = compute_ves_norm(VPoly)
    dL = 0.01

    PtsCross, dLr, ind, N, Rref, VPbis = GG.discretize_vpoly(VPoly, dL,
                                                             D1=None,
                                                             D2=None,
                                                             margin=1.e-9,
                                                             DIn=0., VIn=VIn)
    assert PtsCross.ndim == 2 and PtsCross.shape[1] >= VPoly.shape[1] - 1 and \
           not np.allclose(PtsCross[:, 0], PtsCross[:, -1])
    assert dLr.shape == (PtsCross.shape[1],) and np.all(dLr <= dL)
    assert ind.shape == (PtsCross.shape[1],)
    assert np.all(np.unique(ind) == ind)
    assert np.all(~np.isnan(ind))
    assert np.max(ind) < PtsCross.shape[1]
    assert N.shape == (VPoly.shape[1] - 1,) and np.all(N >= 1)
    assert Rref.shape == (PtsCross.shape[1],)
    assert np.all(Rref == PtsCross[0, :])
    assert VPbis.ndim == 2 and VPbis.shape[1] >= VPoly.shape[1]

    PtsCross, dLr, ind, N, Rref, VPbis = GG.discretize_vpoly(VPoly, dL,
                                                             D1=[0., 2.],
                                                             D2=[-2., 0.],
                                                             margin=1.e-9,
                                                             DIn=0.05, VIn=VIn)
    assert np.all(PtsCross[0, :] >= 0.) and np.all(PtsCross[0, :] <= 2.) and \
           np.all(PtsCross[1, :] >= -2.) and np.all(PtsCross[1, :] <= 0.)
    assert np.all(Path(VPoly.T).contains_points(PtsCross.T))
    assert dLr.shape == (PtsCross.shape[1],) and np.all(dLr <= dL)
    assert ind.shape == (PtsCross.shape[1],) and np.all(np.unique(ind) == ind)
    assert np.all(~np.isnan(ind))
    assert N.shape == (VPoly.shape[1] - 1,) and np.all(N >= 1)
    assert Rref.size > 3 * PtsCross.shape[1]
    assert VPbis.ndim == 2 and VPbis.shape[1] >= VPoly.shape[1]

    PtsCross, dLr, ind, N, Rref, VPbis = GG.discretize_vpoly(VPoly, dL,
                                                             D1=[0., 2.],
                                                             D2=[-2., 0.],
                                                             margin=1.e-9,
                                                             DIn=-0.05,
                                                             VIn=VIn)
    assert np.all(PtsCross[0, :] >= 0. - 0.05)
    assert np.all(PtsCross[0, :] <= 2.)
    assert np.all(PtsCross[1, :] >= -2. - 0.05)
    assert np.all(PtsCross[1, :] <= 0.)
    assert np.all(~Path(VPoly.T).contains_points(PtsCross.T))


# =============================================================================
# Ves  - Vmesh
# =============================================================================
def test03_Ves_Vmesh_Tor():
    r_min_max, z_min_max = compute_min_max_r_and_z(VPoly)
    dR, dZ, dRPhi = 0.05, 0.05, 0.05
    LDPhi = [None,
             [3. * np.pi / 4., 5. * np.pi / 4.],
             [-np.pi / 4., np.pi / 4.]]

    for ii in range(len(LDPhi)):
        out = GG._Ves_Vmesh_Tor_SubFromD_cython(dR, dZ, dRPhi,
                                                r_min_max,
                                                z_min_max,
                                                DR=[0.5, 2.],
                                                DZ=[0., 1.2],
                                                DPhi=LDPhi[ii],
                                                limit_vpoly=VPoly,
                                                out_format='(R,Z,Phi)',
                                                margin=1.e-9)
        pts, vol_res, ind, r_res, z_res, vec_phi_res, sz_r, sz_z = out
        assert pts.ndim == 2 and pts.shape[0] == 3
        assert np.all(pts[0, :] >= 1.) and np.all(pts[0, :] <= 2.), \
            " X coordinates not in right bounds"
        assert np.all(pts[1, :] >= 0.) and np.all(pts[1, :] <= 1.), \
            " Y coordinates not in right bounds"
        marg = np.abs(np.arctan(np.mean(vec_phi_res) / np.min(VPoly[1, :])))
        if LDPhi[ii] is not None:
            LDPhi[ii][0] = np.arctan2(np.sin(LDPhi[ii][0]),
                                      np.cos(LDPhi[ii][0]))
            LDPhi[ii][1] = np.arctan2(np.sin(LDPhi[ii][1]),
                                      np.cos(LDPhi[ii][1]))
            if LDPhi[ii][0] <= LDPhi[ii][1]:
                assert np.all((pts[2, :] >= LDPhi[ii][0] - marg) &
                              (pts[2, :] <= LDPhi[ii][1] + marg))
            else:
                assert np.all((pts[2, :] >= LDPhi[ii][0] - marg) |
                              (pts[2, :] <= LDPhi[ii][1] + marg))
        assert vol_res.shape == (pts.shape[1],)
        lc = [
            ind.shape == (pts.shape[1],),
            'int' in ind.dtype.name,
            np.unique(ind).size == ind.size,
            np.all(ind == np.unique(ind)),
            np.all(ind >= 0),
        ]
        if not all(lc):
            msg = str(lc)
            raise Exception(msg)
        assert vec_phi_res.ndim == 1

        out = GG._Ves_Vmesh_Tor_SubFromInd_cython(dR, dZ, dRPhi,
                                                  r_min_max, z_min_max, ind,
                                                  Out='(R,Z,Phi)',
                                                  margin=1.e-9)
        Ptsi, dVi, dRri, dZri, dRPhiri = out
        assert np.allclose(pts, Ptsi)
        assert np.allclose(vol_res, dVi)
        assert r_res == dRri and z_res == dZri
        assert np.allclose(vec_phi_res, dRPhiri)


def test04_ves_vmesh_lin():
    XMinMax = np.array([0., 10.])
    YMinMax = np.array([np.min(VPoly[0, :]), np.max(VPoly[0, :])])
    ZMinMax = np.array([np.min(VPoly[1, :]), np.max(VPoly[1, :])])
    dX, dY, dZ = 0.05, 0.05, 0.05

    out = GG._Ves_Vmesh_Lin_SubFromD_cython(dX, dY, dZ, XMinMax,
                                            YMinMax, ZMinMax,
                                            DX=[8., 15.],
                                            DY=[0.5, 2.],
                                            DZ=[0., 1.2],
                                            limit_vpoly=VPoly,
                                            margin=1.e-9)
    Pts, dV, ind, dXr, dYr, dZr = out
    assert Pts.ndim == 2 and Pts.shape[0] == 3
    assert np.all(Pts[0, :] >= 8.) and np.all(Pts[0, :] <= 10.) and \
           np.all(Pts[1, :] >= 1.) and np.all(Pts[1, :] <= 2.) and \
           np.all(Pts[2, :] >= 0.) and np.all(Pts[2, :] <= 1.)
    assert all([
        ind.shape == (Pts.shape[1],),
        'int' in ind.dtype.name,
        np.unique(ind).size == ind.size,
        np.all(ind == np.unique(ind)), np.all(ind >= 0),
    ])

    out = GG._Ves_Vmesh_Lin_SubFromInd_cython(dX, dY, dZ,
                                              XMinMax, YMinMax, ZMinMax,
                                              ind, margin=1.e-9)
    Ptsi, dVi, dXri, dYri, dZri = out
    assert np.allclose(Pts, Ptsi)
    assert np.allclose(dV, dVi)
    assert dXr == dXri and dYr == dYri and dZr == dZri


# =============================================================================
# Ves  - Solid angles
# =============================================================================


def test05_sa_integ_map(ves_poly=VPoly, debug=0):
    import matplotlib.pyplot as plt
    print()

    lblock = [False, True]
    lstep_rz = [0.015, 0.02]

    for (block, step_rz) in zip(lblock, lstep_rz):

        config = tf.load_config("A1")
        kwdargs = config.get_kwdargs_LOS_isVis()
        ves_poly = kwdargs["ves_poly"]

        # coordonnées en x,y,z:
        part = np.array([[3, 0.1, 0.1],
                         [3, 0.1, 0.]], order='F').T
        part_rad = np.r_[0.001,
                         0.0012]
        rstep = zstep = step_rz
        phistep = 0.08
        limits_r, limits_z = compute_min_max_r_and_z(ves_poly)
        DR = None
        DZ = None
        DPhi = [-0.1, 0.1]

        # -- Getting cython APPROX computation --------------------------------
        res = GG.compute_solid_angle_map(part, part_rad,
                                         rstep, zstep, phistep,
                                         limits_r, limits_z,
                                         DR=DR, DZ=DZ,
                                         DPhi=DPhi,
                                         block=block,
                                         limit_vpoly=ves_poly,
                                         **kwdargs,
                                         )
        pts, sa_map_cy, ind, reso_r_z = res

        # check sizes
        npts_ind = np.size(ind)
        dim, npts = np.shape(pts)
        npts_sa, sz_p = np.shape(sa_map_cy)

        if debug > 2:
            print(f"sa_map_cy is of size {npts_sa},{sz_p}")

        # Checking shapes, sizes, types
        assert dim == 2
        assert sz_p == np.shape(part)[1]
        assert npts_ind == npts
        assert npts == npts_sa
        assert isinstance(reso_r_z, float)

        # -- Getting cython EXACT computation ---------------------------------
        kwdargs = config.get_kwdargs_LOS_isVis()
        res = GG.compute_solid_angle_map(part, part_rad,
                                         rstep, zstep, phistep,
                                         limits_r, limits_z,
                                         approx=False,
                                         DR=DR, DZ=DZ,
                                         DPhi=DPhi,
                                         block=block,
                                         **kwdargs,
                                         limit_vpoly=ves_poly
                                         )
        pts_ex, sa_map_cy_ex, ind_ex, reso_r_z_ex = res

        if debug > 2:
            print(f"sa_map_cy_ex is of size {npts_sa},{sz_p}")

        # Checking if pts, indices and resolution same as with approx
        assert np.allclose(pts_ex, pts)
        assert np.allclose(ind_ex, ind)
        assert reso_r_z_ex == reso_r_z

        # ... Testing with python function ....................................
        res = GG._Ves_Vmesh_Tor_SubFromD_cython(rstep, zstep, phistep,
                                                limits_r, limits_z,
                                                limit_vpoly=ves_poly,
                                                DR=DR, DZ=DZ,
                                                DPhi=DPhi,
                                                out_format='(X,Y,Z)',
                                                margin=1.e-9)

        pts_xyz, dvol, ind_disc, reso_r, reso_z, reso_phi, sz_r, sz_z = res

        npts_disc = np.shape(pts_xyz)[1]

        sang = config.calc_solidangle_particle(pts_xyz,
                                               part,
                                               part_rad,
                                               block=block,
                                               approx=True)

        sang_ex = config.calc_solidangle_particle(pts_xyz,
                                                  part,
                                                  part_rad,
                                                  block=block,
                                                  approx=False)

        pts_disc = GG.coord_shift(pts_xyz, in_format='(X,Y,Z)',
                                  out_format='(R,Z)')

        assert (npts_disc, sz_p) == np.shape(sang)
        assert npts_disc >= npts
        assert reso_r * reso_z == reso_r_z
        if ves_poly is None:
            assert npts_sa == sz_z * sz_r, f"sizes r and z = {sz_r}{sz_z}"

        sa_map_py = np.zeros((npts_sa, sz_p))
        sa_map_py_ex = np.zeros((npts_sa, sz_p))

        r0 = limits_r[0]
        z0 = limits_z[0]

        lvisited = []
        for ii in range(npts_disc):
            i_r = int(np.abs(r0 - pts_disc[0, ii]) // reso_r)
            i_z = int(np.abs(z0 - pts_disc[1, ii]) // reso_z)
            ind_pol = int(i_r * sz_z + i_z)
            if ves_poly is not None:
                wh_eq = np.where(ind == ind_pol)
                ind_pol_u = wh_eq[0][0]
            else:
                ind_pol_u = ind_pol
            if ind_pol_u not in lvisited:
                lvisited.append(ind_pol_u)
            for pp in range(sz_p):
                sa_map_py[ind_pol_u, pp] += sang[ii, pp] * reso_phi[i_r]
                sa_map_py_ex[ind_pol_u, pp] += sang_ex[ii, pp] * reso_phi[i_r]

        if debug > 0:
            for pp in range(sz_p):
                fig = plt.figure(figsize=(14, 8))
                ax = plt.subplot(121)
                # import pdb; pdb.set_trace()
                ax.scatter(pts[0, :], pts[1, :],
                           marker="s", edgecolors="None",
                           s=10, c=sa_map_cy[:, pp].flatten(),
                           vmin=0, vmax=max(sa_map_cy[:, pp].max(),
                                            sa_map_py[:, pp].max()))
                ax.set_title("cython function")
                ax = plt.subplot(122)
                im = ax.scatter(pts[0, :], pts[1, :],
                                marker="s", edgecolors="None",
                                s=10, c=sa_map_py[:, pp].flatten(),
                                vmin=0, vmax=max(sa_map_cy[:, pp].max(),
                                                 sa_map_py[:, pp].max()))
                fig.colorbar(im, ax=ax)
                ax.set_title("python reconstruction")
                plt.savefig("comparaison_part"+str(pp))

        max_py = np.max(sa_map_py)
        max_cy = np.max(sa_map_cy)
        max_py_ex = np.max(sa_map_py_ex)
        max_cy_ex = np.max(sa_map_cy_ex)

        err = np.abs(sa_map_py - sa_map_cy) / max(max_py, max_cy)
        print("max error approx py vs cy =", np.max(err))

        err = np.abs(sa_map_py_ex - sa_map_cy_ex) / max(max_py_ex, max_cy_ex)
        print("max error exacts py vs cy =", np.max(err))

        err = np.abs(sa_map_py_ex - sa_map_py) / max(max_py, max_py_ex)
        print("max error python approx vs exact =", np.max(err))

        err = np.abs(sa_map_cy_ex - sa_map_cy) / max(max_cy, max_cy_ex)
        print("max error cython approx vs exact =", np.max(err))

        assert np.allclose(sa_map_cy, sa_map_py, atol=1e-16, rtol=1e-16,
                           equal_nan=True)
        assert np.allclose(sa_map_cy_ex, sa_map_py_ex, atol=1e-16, rtol=1e-16,
                           equal_nan=True)
        assert np.allclose(sa_map_cy, sa_map_cy_ex, atol=1e-14, rtol=1e-16,
                           equal_nan=True)
        assert np.allclose(sa_map_py, sa_map_py_ex, atol=1e-14, rtol=1e-16,
                           equal_nan=True)

    # ...
    return


# -*- coding: utf-8 -*-


import warnings


import numpy as np
import numpy.polynomial.polynomial as npoly
import scipy.optimize as scpopt
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


# ##############################################################
# ##############################################################
#           Finding reflection points
# ##############################################################


def _get_pts2pt(
    coll=None,
    key=None,
):

    # ---------
    # key

    key, cls = coll.get_optics_cls(optics=key)
    key, cls = key[0], cls[0]
    dgeom = coll.dobj[cls][key]['dgeom']

    lcls = ['crystal', 'grating']
    if cls not in lcls:
        msg = (
            "Wrong class for reflections:\n"
            f"\t -allowed: {lcls}\n"
            f"\t -cls: {cls}\n"
        )
        raise Exception(msg)

    # -------------------
    #     Planar
    # -------------------

    if dgeom['type'] == 'planar':

        def pts2pt(
            pt_x=None,
            pt_y=None,
            pt_z=None,
            # pts
            pts_x=None,
            pts_y=None,
            pts_z=None,
            # surface
            cent=dgeom['cent'],
            nin=dgeom['nin'],
            e0=dgeom['e0'],
            e1=dgeom['e1'],
            # return
            strict=None,
            return_xyz=None,
            return_x01=None,
            debug=None,
            # timing
            dt=None,
        ):
            """

            cf. Notes_Upgrades/raytracing_surface3draytracing_surface3d.pdf

            """

            # ------------------------------------------
            # Get local coordinates of reflection points

            # get parameters for k
            CA = np.r_[
                pt_x - cent[0], pt_y - cent[1], pt_z - cent[2],
            ]
            CAdotn = np.sum(CA * nin)

            ABx = pts_x - pt_x
            ABy = pts_y - pt_y
            ABz = pts_z - pt_z

            length = np.sqrt(ABx**2 + ABy**2 + ABz**2)

            # get k
            kk = CAdotn / (
                2*CAdotn + (nin[0] * ABx + nin[1]*ABy + nin[2]*ABz)
            )

            # get E
            Ex = pt_x + kk * ABx
            Ey = pt_y + kk * ABy
            Ez = pt_z + kk * ABz

            # x0, x1
            x0 = (
                (Ex - cent[0])*e0[0]
                + (Ey - cent[1])*e0[1]
                + (Ez - cent[2])*e0[2]
            )
            x1 = (
                (Ex - cent[0])*e1[0]
                + (Ey - cent[1])*e1[1]
                + (Ez - cent[2])*e1[2]
            )

            if return_xyz:
                # get D
                Dx = cent[0] + x0 * e0[0] + x1*e1[0]
                Dy = cent[1] + x0 * e0[1] + x1*e1[1]
                Dz = cent[2] + x0 * e0[2] + x1*e1[2]

            if return_xyz and return_x01:
                return Dx, Dy, Dz, x0, x1
            elif return_xyz:
                return Dx, Dy, Dz
            elif return_x01:
                return x0, x1

    # ----------------
    #   Cylindrical
    # ----------------

    elif dgeom['type'] == 'cylindrical':

        iplan = np.isinf(dgeom['curve_r']).nonzero()[0][0]
        eax = ['e0', 'e1'][iplan]
        erot = ['e0', 'e1'][1-iplan]

        rc = dgeom['curve_r'][1 - iplan]
        rcs = np.sign(rc)
        rca = np.abs(rc)

        # pick solver
        if rcs > 0:
            solver = root_cyl_concave
        else:
            solver = root_cyl_convex


        def pts2pt(
            pt_x=None,
            pt_y=None,
            pt_z=None,
            # pts
            pts_x=None,
            pts_y=None,
            pts_z=None,
            # surface
            O=dgeom['cent'] + dgeom['nin'] * rc,
            rcs=rcs,
            rca=rca,
            eax=dgeom[eax],
            erot=dgeom[erot],
            iplan=iplan,
            # limits
            thetamax=dgeom['extenthalf'][1-iplan],
            xmax=dgeom['extenthalf'][iplan],
            # local coordinates
            nin=dgeom['nin'],
            # solver
            solver=solver,
            # return
            strict=True,
            return_xyz=None,
            return_x01=None,
            # number of k for interpolation
            nk=None,
            debug=None,
            # timing
            dt=None,
        ):
            """

            cf. Notes_Upgrades/raytracing_surface3draytracing_surface3d.pdf

            """

            # ------------------------------------------
            # Get local coordinates of reflection points

            A = np.r_[pt_x, pt_y, pt_z]
            B = np.r_[pts_x[0], pts_y[0], pts_z[0]]
            Dx, Dy, Dz, xx, theta = _common_prepare(nb=pts_x.size)

            if strict:
                xmargin = 1.
                thetatot = thetamax
            else:
                xmargin = 1. + 0.01
                thetatot = thetamax*(1. + 0.01)

            # defining zones for kk (point , plane unit vector)
            em = np.cos(thetatot)*(erot) + rcs * np.sin(thetatot)*(-nin)
            eM = np.cos(thetatot)*(erot) - rcs * np.sin(thetatot)*(-nin)

            lzones = [
                (
                    (O - (xmax*xmargin)*eax, eax),
                    (O + (xmax*xmargin)*eax, -eax),
                    (O, em),
                    (O, -eM),
                ),

            ]

            if rcs > 0:
                lzones.append((
                    (O - (xmax*xmargin)*eax, eax),
                    (O + (xmax*xmargin)*eax, -eax),
                    (O, -em),
                    (O, eM),
                ))

            # ------------
            # loop on pts

            for ii in range(pts_x.size):
                B[:] = pts_x[ii], pts_y[ii], pts_z[ii]
                AB = B - A
                eAB = AB / np.linalg.norm(AB)

                # solver for root finding
                roots = solver(
                    lzones=lzones,
                    # points
                    O=O,
                    A=A,
                    B=B,
                    # radius
                    rcs=rcs,
                    rca=rca,
                    # unit vectors
                    nin=nin,
                    eax=eax,
                    # options
                    nk=nk,
                    # timing
                    dt=dt,
                    # debug
                    debug=debug and ii == 0,
                )

                if roots is None:
                    continue

                # else:
                #     continue

                # nin, xx, D
                (
                    nix, niy, niz,
                    Dxi, Dyi, Dzi, xxi,
                ) = _get_Dnin_from_k_cyl(
                    O=O, eax=eax,
                    Ax=pt_x, Ay=pt_y, Az=pt_z,
                    Bx=pts_x[ii], By=pts_y[ii], Bz=pts_z[ii],
                    kk=roots,
                    rcs=rcs,
                    rca=rca,
                    nin=nin,
                )

                # theta, xx
                thetai = rcs * np.arctan2(
                    -(nix*erot[0] + niy*erot[1] + niz*erot[2]),
                    nix*nin[0] + niy*nin[1] + niz*nin[2],
                )

                # ------ DEBUG -----------------------------
                if debug and ii == 0:
                    _debug_new(**locals())
                # -----------------------------------------

                if strict is True:
                    if np.abs(xxi) > xmax or np.abs(thetai) > thetamax:
                        continue

                Dx[ii], Dy[ii], Dz[ii] = Dxi, Dyi, Dzi
                xx[ii], theta[ii] = xxi, thetai

            # safety check
            iok = np.isfinite(Dx)
            Dx = Dx[iok]
            Dy = Dy[iok]
            Dz = Dz[iok]
            theta = theta[iok]
            xx = xx[iok]

            # TODO : find a way to hanbdle complex (self-intersection polygons)
            # Exceptionally a polygon (xx, theta) can self-intersect

            # return
            if return_xyz and return_x01:
                if iplan == 0:
                    return Dx, Dy, Dz, xx, theta
                else:
                    return Dx, Dy, Dz, theta, xx
            elif return_xyz:
                return Dx, Dy, Dz
            elif return_x01:
                if iplan == 0:
                    return xx, theta
                else:
                    return theta, xx

    # ----------------
    #   Spherical
    # ----------------

    elif dgeom['type'] == 'spherical':

        rc = dgeom['curve_r'][0]
        rcs = np.sign(rc)
        rca = np.abs(rc)
        
        # pick solver
        if rcs > 0:
            solver = root_sph_concave
        else:
            solver = root_sph_convex

        def pts2pt(
            pt_x=None,
            pt_y=None,
            pt_z=None,
            # pts
            pts_x=None,
            pts_y=None,
            pts_z=None,
            # surface
            O=dgeom['cent'] + rc*dgeom['nin'],
            rcs=rcs,
            rca=rca,
            # limits
            dthetamax=dgeom['extenthalf'][1],
            phimax=dgeom['extenthalf'][0],
            # local coordinates
            nin=dgeom['nin'],
            e0=dgeom['e0'],
            e1=dgeom['e1'],
            # solving
            nk=None,
            solver=solver,
            # return
            strict=None,
            return_xyz=None,
            return_x01=None,
            debug=None,
            # timing
            dt=None,
        ):
            """

            cf. Notes_Upgrades/raytracing_surface3draytracing_surface3d.pdf

            """

            # ------------------------------------------
            # Get local coordinates of reflection points

            # get parameters for k
            A = np.r_[pt_x, pt_y, pt_z]
            B = np.r_[pts_x[0], pts_y[0], pts_z[0]]
            Dx, Dy, Dz, phi, dtheta = _common_prepare(nb=pts_x.size)

            if strict:
                phitot = phimax
                dthetatot = dthetamax                
            else:
                phitot = phimax*(1. + 0.01)
                dthetatot = dthetamax*(1. + 0.01)

             # defining zones for kk (point , plane unit vector)
            epm = np.cos(phitot)*e0 + rcs * np.sin(phitot)*(-nin)
            epM = np.cos(phitot)*e0 - rcs * np.sin(phitot)*(-nin)

            etm = np.cos(dthetatot)*e1 + rcs * np.sin(dthetatot)*(-nin)
            etM = np.cos(dthetatot)*e1 - rcs * np.sin(dthetatot)*(-nin)

            lzones = [
                (
                    (O, epm),
                    (O, -epM),
                    (O, etm),
                    (O, -etM),
                ),
            ]

            if rcs > 0:
                lzones.append((
                    (O, -epm),
                    (O, epM),
                    (O, -etm),
                    (O, etM),
                ))

            # loop on pts
            for ii in range(pts_x.size):
                B[:] = pts_x[ii], pts_y[ii], pts_z[ii]
                AB = B - A
                eAB = AB / np.linalg.norm(AB)
                
                # solver for root finding
                roots = solver(
                    lzones=lzones,
                    # points
                    O=O,
                    A=A,
                    B=B,
                    # radius
                    rcs=rcs,
                    rca=rca,
                    # unit vectors
                    nin=nin,
                    # options
                    nk=nk,
                    # timing
                    dt=dt,
                    # debug
                    debug=debug and ii == 0,
                )
                
                if roots is None:
                    continue

                # nin, xx, D
                (
                    nix, niy, niz,
                    Dxi, Dyi, Dzi,
                ) = _get_Dnin_from_k_sph(
                    O=O,
                    Ax=pt_x, Ay=pt_y, Az=pt_z,
                    Bx=pts_x[ii], By=pts_y[ii], Bz=pts_z[ii],
                    kk=roots,
                    rcs=rcs,
                    rca=rca,
                    nin=nin,
                )

                # local coordinates
                dthetai = - rcs * np.arcsin(nix*e1[0] + niy*e1[1] + niz*e1[2])
                phii = - rcs * np.arcsin(
                    (nix*e0[0] + niy*e0[1] + niz*e0[2]) / np.cos(dthetai)
                )
                
                # ------ DEBUG -----------------------------
                if debug and ii == 0:
                    _debug_new(**locals())
                # -----------------------------------------

                if strict is True:
                    if np.abs(phii) > phimax or np.abs(dthetai) > dthetamax:
                        continue

                Dx[ii], Dy[ii], Dz[ii] = Dxi, Dyi, Dzi
                phi[ii], dtheta[ii] = phii, dthetai

            # safety check
            iok = np.isfinite(Dx)
            Dx = Dx[iok]
            Dy = Dy[iok]
            Dz = Dz[iok]
            dtheta = dtheta[iok]
            phi = phi[iok]

            # TODO : find a way to hanbdle complex (self-intersection polygons)
            # Exceptionally a polygon (xx, theta) can self-intersect

            # return            
            if return_xyz and return_x01:
                return Dx, Dy, Dz, phi, dtheta
            elif return_xyz:
                return Dx, Dy, Dz
            elif return_x01:
                return phi, dtheta

    # ----------------
    #   Toroidal
    # ----------------

    elif dgeom['type'] == 'toroidal':

        raise NotImplementedError()

    return pts2pt


# #################################################################
# #################################################################
#           Common formulas
# #################################################################


# Not used
def _mindist_2lines(A=None, B=None, O=None, eax=None):
    OAe = np.cross(A - O, eax)
    ABe = np.cross(B - A, eax)
    OAe2 = np.sum(OAe**2)
    ABe2 = np.sum(ABe**2)
    OAeABe = np.sum(OAe * ABe)
    kk = -OAeABe/ABe2
    return kk, OAe2 - OAeABe**2/ABe2


def _kminmax_ptinz(pt=None, lp=None):
    return np.all([
        np.sum((pt - oo)*ep) > 0 for oo, ep in lp
    ])


def _kminmax_ptinz2(pts=None, lp=None):
    return np.all(
        [
            np.sum((pts - oo[:, None])*ep[:, None], axis=0) > 0
            for oo, ep in lp
        ],
        axis=0,
    )


def _kminmax_plane(
    A=None,
    B=None,
    lzones=None,
    kmin=0,
    kmax=1,
    nin=None,
    # options
    nk=None,
    return_vector=True,
    # timing
    dt=None,
):

    # ---------------
    # check inputs

    if nk is None:
        nk = 100

    AB = B - A

    # ---------------
    # get lkin, lkout

    # Intersection with planes (same planes for all zones)
    lk = []
    for jj, (oo, ep) in enumerate(lzones[0]):
        AO = oo - A
        AOe = np.sum(AO*ep)
        ABe = np.sum(AB * ep)

        if np.abs(ABe) < 1e-16:
            continue
        kk = AOe / ABe
        if kk < 0 or kk > 1:
            continue
        lk.append(kk)

    kk = np.r_[0., np.sort(lk) + 1.e-13, 1.]

    kin = np.array([
        _kminmax_ptinz2(
            pts=A[:, None] + kk[None, :]*AB[:, None],
            lp=lp,
        )
        for lp in lzones
    ])

    iok = np.any(kin, axis=1)

    # trivial case 1
    if not np.any(iok):
        return None

    kin = kin[iok, :]

    # trivial case 2
    if np.any(np.all(kin, axis=1)):
        if return_vector:
            return np.linspace(0, 1, nk)
        else:
            return (0, 1)

    # non-trivial cases
    iin = kin.nonzero()[1]
    iout = kin.shape[1] - kin[:, ::-1].nonzero()[1]

    if iin.shape[0] > 1:
        msg = (
            "In _kminmax_plane(), identified case with iin.shape[0] > 1\n"
            f"kin: {kin}\n"
            f"iin: {iin}\n"
        )
        warnings.warn(msg)
        # import pdb; pdb.set_trace()     # DB

    # remove simulatenous in/out
    idel = [ii for ii in range(iin.size) if iin[ii] == iout[ii]]
    if len(idel) > 0:
        iin = np.delete(iin, idel)
        iout = np.delete(iout, idel)

    # concatenate simultaneous out/in
    if iin.size > 1:
        idel = [ii for ii in range(1, iin.size) if iin[ii] == iout[ii-1]]
        if len(idel) > 0:
            iin = np.delete(iin, idel)
            iout = np.delete(iout, np.array(idel)-1)

    # -----------
    # return

    if iin.size == 0:
        if return_vector:
            return np.linspace(0, 1, nk)
        else:
            return (0, 1)

    elif iin.size == 1:
        if return_vector:
            return np.linspace(kk[iin[0]], kk[iout[0]], nk)
        else:
            return (kk[iin[0]], kk[iout[0]])

    else:
        if not return_vector:
            msg = "multiple brackets!"
            raise Exception(msg)

        return np.concatenate(
            tuple([
                np.linspace(kk[iin[ii]], kk[iout[ii]], nk)
                for ii in range(iin.size)
            ])
        )


def _get_Dnin_from_k_cyl(
    O=None, eax=None,
    Ax=None, Ay=None, Az=None,
    Bx=None, By=None, Bz=None,
    kk=None,
    rcs=None,
    rca=None,
    nin=None,
    # debug
    debug=None,
):
    Ex = Ax + kk*(Bx - Ax)
    Ey = Ay + kk*(By - Ay)
    Ez = Az + kk*(Bz - Az)

    xx = (Ex - O[0])*eax[0] + (Ey - O[1])*eax[1] + (Ez - O[2])*eax[2]
    nix = (Ex - O[0]) - xx*eax[0]
    niy = (Ey - O[1]) - xx*eax[1]
    niz = (Ez - O[2]) - xx*eax[2]
    ninorm = np.sqrt(nix**2 + niy**2 + niz**2)
    nix = - rcs * nix / ninorm
    niy = - rcs * niy / ninorm
    niz = - rcs * niz / ninorm

    # handle 2 zones
    sign = np.sign(nix*nin[0] + niy*nin[1] + niz*nin[2])
    if rcs > 0:
        nix = sign*nix
        niy = sign*niy
        niz = sign*niz
    else:
        assert np.all(sign > 0.)

    Dx = O[0] + xx*eax[0] - rcs * rca*nix
    Dy = O[1] + xx*eax[1] - rcs * rca*niy
    Dz = O[2] + xx*eax[2] - rcs * rca*niz
    return nix, niy, niz, Dx, Dy, Dz, xx


def _get_Dnin_from_k_sph(
    O=None,
    Ax=None, Ay=None, Az=None,
    Bx=None, By=None, Bz=None,
    kk=None,
    rcs=None,
    rca=None,
    nin=None,
):
    Ex = Ax + kk*(Bx - Ax)
    Ey = Ay + kk*(By - Ay)
    Ez = Az + kk*(Bz - Az)

    ninorm = np.sqrt((Ex - O[0])**2 + (Ey - O[1])**2 + (Ez - O[2])**2)
    nix = - rcs * (Ex - O[0]) / ninorm
    niy = - rcs * (Ey - O[1]) / ninorm
    niz = - rcs * (Ez - O[2]) / ninorm

    sign = np.sign(nix*nin[0] + niy*nin[1] + niz*nin[2])
    if rcs > 0:
        nix = sign*nix
        niy = sign*niy
        niz = sign*niz
    else:
        assert np.all(sign > 0.)

    Dx = O[0] - rcs * rca*nix
    Dy = O[1] - rcs * rca*niy
    Dz = O[2] - rcs * rca*niz
    return nix, niy, niz, Dx, Dy, Dz


def _get_DADB(
    Ax=None, Ay=None, Az=None,
    Bx=None, By=None, Bz=None,
    Dx=None, Dy=None, Dz=None,
    nix=None, niy=None, niz=None,
):

    DA = np.sqrt((Ax - Dx)**2 + (Ay - Dy)**2 + (Az - Dz)**2)
    DB = np.sqrt((Bx - Dx)**2 + (By - Dy)**2 + (Bz - Dz)**2)

    DAn = (Ax - Dx)*nix + (Ay - Dy)*niy + (Az - Dz)*niz
    DBn = (Bx - Dx)*nix + (By - Dy)*niy + (Bz - Dz)*niz
    return DB/DBn - DA/DAn


def _common_prepare(nb=None):
    return (
        np.full((nb,), np.nan),
        np.full((nb,), np.nan),
        np.full((nb,), np.nan),
        np.full((nb,), np.nan),
        np.full((nb,), np.nan),
    )


# #################################################################
# #################################################################
#           Root finding - Cylindrical
# #################################################################


def root_cyl_concave(
    lzones=None,
    # points
    O=None,
    A=None,
    B=None,
    # radius
    rcs=None,
    rca=None,
    # unit vectors
    nin=None,
    eax=None,
    # options
    nk=None,
    # nrobust=3,
    nrobust=10,
    # timing
    dt=None,
    # debug
    debug=None,
    # unused
    **kwdargs,
):

    # compute kmin, kmax
    kk = _kminmax_plane(
        A=A,
        B=B,
        lzones=lzones,
        nin=nin,
        nk=nk,
        # timing
        dt=dt,
    )

    if kk is None:
        return

    # nin
    (
        nix, niy, niz,
        Dxi, Dyi, Dzi,
        # D1x, D1y, D1z,
        xxi,
    ) = _get_Dnin_from_k_cyl(
        O=O, eax=eax,
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        kk=kk,
        rcs=rcs,
        rca=rca,
        nin=nin,
    )

    # derive DAn, DB, DBn, DA
    eq = _get_DADB(
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        Dx=Dxi, Dy=Dyi, Dz=Dzi,
        nix=nix, niy=niy, niz=niz,
    )
    iok = np.isfinite(eq)
    kk = kk[iok]
    eq = eq[iok]

    # find indices of sign changes
    i0 = (eq[:-1] * eq[1:] < 0).nonzero()[0]
    if len(i0) == 1:

        # concave => harder
        # indices used for linear fit
        ind = np.arange(max(0, i0-nrobust), min(i0+nrobust, kk.size))
        kint = [kk[ind][0], kk[ind][-1]]

        # fit parabola
        out = npoly.Polynomial.fit(
            kk[ind],
            eq[ind],
            deg=3,
            domain=kint,
            window=kint,
        )

        # fit
        line = out(kk[ind])


        # linear least square fit
        # xx_m = np.mean(kk[ind])
        # yy_m = np.mean(eq[ind])
        # xx2_m = np.mean(kk[ind]**2)
        # xy_m = np.mean(kk[ind] * eq[ind])

        # coefs
        # aa = (xy_m - xx_m*yy_m) / (xx2_m - xx_m**2)
        # bb = yy_m - aa * xx_m

        # fit and threshold
        # line = aa*kk[ind] + bb
        thr = abs(line[0] - line[-1]) * 0.02
        if np.any(np.abs(eq[ind] - line) > thr):
            return

        # roots
        roots = out.roots()
        roots = np.real(roots[np.isreal(roots)])
        roots = [
            rr for rr in roots
            if rr >= kint[0] and rr <= kint[1]
        ]

    else:
        return

    # ----------------
    # debug

    if debug:
        plt.figure()
        plt.plot(
            kk, eq, 'x-k',
            kk[ind], line, 'o-r',
            kk[ind], line + thr, '--r',
            kk[ind], line - thr, '--r',
            [kk[i0]], [eq[i0]], 'xr',
        )
        plt.axhline(0, c='k', ls='--')
        plt.axvline(roots[0], c='k', ls='--')
        plt.gca().set_title(f'thr = {thr}')
        print(kint, roots)

    return roots[0]


def _func(
    kk,
    O=None,
    A=None,
    B=None,
    rcs=None,
    rca=None,
    nin=None,
    eax=None,
):

    (
        nix, niy, niz,
        Dxi, Dyi, Dzi,
        # D1x, D1y, D1z,
        xxi,
    ) = _get_Dnin_from_k_cyl(
        O=O, eax=eax,
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        kk=kk,
        rcs=rcs,
        rca=rca,
        nin=nin,
    )

    return _get_DADB(
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        Dx=Dxi, Dy=Dyi, Dz=Dzi,
        nix=nix, niy=niy, niz=niz,
    )


def root_cyl_convex(
    # points
    O=None,
    A=None,
    B=None,
    # radius
    rcs=None,
    rca=None,
    # unit vectors
    nin=None,
    eax=None,
    # options
    lzones=None,
    # timing
    dt=None,
    # unused
    **kwdargs,
):

    # get bracket
    bracket = _kminmax_plane(
        A=A,
        B=B,
        lzones=lzones,
        nin=nin,
        # options
        nk=None,
        return_vector=False,
        # timing
        dt=dt,
    )

    # solve
    roots = scpopt.root_scalar(
        _func,
        args=(O, A, B, rcs, rca, nin, eax),
        method='brentq',    # 'bisect', 'brentq'
        bracket=bracket,
        fprime=None,
        fprime2=None,
        x0=None,
        x1=None,
        xtol=1e-10,
        rtol=1e-10,
        maxiter=100,
        options=None,
    )

    if roots.converged is not True:
        return
    else:
        return roots.root

# #################################################################
# #################################################################
#           Root finding - Spherical
# #################################################################


def root_sph_concave(
    lzones=None,
    # points
    O=None,
    A=None,
    B=None,
    # radius
    rcs=None,
    rca=None,
    # unit vectors
    nin=None,
    # options
    nk=None,
    # nrobust=3,
    nrobust=10,
    # timing
    dt=None,
    # debug
    debug=None,
    # unused
    **kwdargs,
):

    # compute kmin, kmax
    kk = _kminmax_plane(
        A=A,
        B=B,
        lzones=lzones,
        nin=nin,
        nk=nk,
        # timing
        dt=dt,
    )

    if kk is None:
        return

    # nin, xx, D
    (
        nix, niy, niz,
        Dxi, Dyi, Dzi,
    ) = _get_Dnin_from_k_sph(
        O=O,
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        kk=kk,
        rcs=rcs,
        rca=rca,
        nin=nin,
    )

    # derive DAn, DB, DBn, DA
    eq = _get_DADB(
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        Dx=Dxi, Dy=Dyi, Dz=Dzi,
        nix=nix, niy=niy, niz=niz,
    )
    iok = np.isfinite(eq)
    kk = kk[iok]
    eq = eq[iok]

    # find indices of sign changes
    i0 = (eq[:-1] * eq[1:] < 0).nonzero()[0]
    if len(i0) == 1:
        # indices used for fit
        ind = np.arange(max(0, i0-nrobust), min(i0+nrobust, kk.size))
        kint = [kk[ind][0], kk[ind][-1]]

        # fit parabola
        out = npoly.Polynomial.fit(
            kk[ind],
            eq[ind],
            deg=3,
            domain=kint,
            window=kint,
        )
        
        # fit
        line = out(kk[ind])
            
        thr = abs(line[0] - line[-1]) * 0.02
        if np.any(np.abs(eq[ind] - line) > thr):
            return
        
        # roots
        roots = out.roots()
        roots = np.real(roots[np.isreal(roots)])
        roots = [
            rr for rr in roots
            if rr >= kint[0] and rr <= kint[1]
        ]
    
    else:
        return

    # ----------------
    # debug

    if debug:
        plt.figure()
        plt.plot(
            kk, eq, 'x-k',
            kk[ind], line, 'o-r',
            kk[ind], line + thr, '--r',
            kk[ind], line - thr, '--r',
            [kk[i0]], [eq[i0]], 'xr',
        )
        plt.axhline(0, c='k', ls='--')
        plt.axvline(roots[0], c='k', ls='--')
        plt.gca().set_title(f'thr = {thr}')
        print(kint, roots)

    return roots[0]



# #################################################################
# #################################################################
#           Debug
# #################################################################


def _debug_new(
    # points
    O=None,
    A=None,
    B=None,
    Dxi=None,
    Dyi=None,
    Dzi=None,
    # unit vectors
    nin=None,
    erot=None,
    eax=None,
    nix=None,
    niy=None,
    niz=None,
    # others
    rcs=None,
    rca=None,
    ii=None,
    roots=None,
    xxi=None,
    thetai=None,
    # unused
    **kwdargs,
):

    Ex = A[0] + roots*(B[0] - A[0])
    Ey = A[1] + roots*(B[1] - A[1])
    Ez = A[2] + roots*(B[2] - A[2])

    eqi = _get_DADB(
        Ax=A[0], Ay=A[1], Az=A[2],
        Bx=B[0], By=B[1], Bz=B[2],
        Dx=Dxi, Dy=Dyi, Dz=Dzi,
        nix=nix, niy=niy, niz=niz,
    )

    DA = np.r_[A[0] - Dxi, A[1] - Dyi, A[2] - Dzi]
    DB = np.r_[B[0] - Dxi, B[1] - Dyi, B[2] - Dzi]
    DE = np.r_[Ex - Dxi, Ey - Dyi, Ez - Dzi]
    DAn = np.linalg.norm(DA)
    DBn = np.linalg.norm(DB)
    DEn = np.linalg.norm(DE)
    cosA = np.sum(DE * DA) / (DAn * DEn)
    cosB = np.sum(DE * DB) / (DBn * DEn)
    print()
    print(f'\nii = {ii}', xxi, thetai)
    print('rcs', rcs)
    print('rca', rca)
    print('nix, niy, niz', nix, niy, niz)
    print('erot', erot)
    print('nin', nin)
    print('eax', eax)
    print('O', O)
    print('E', Ex, Ey, Ez)
    print('OE', Ex - O[0], Ey - O[1], Ez - O[2])
    print('roots', roots)
    print('eq', eqi)
    print('cosA vs cosB', cosA, cosB)

    # 3d plot
    fig = plt.figure()
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection='3d')
    ax.set_title(f"kk = {roots}")
    ax.plot(
        np.r_[O[0], O[0] - rcs*rca*nin[0]],
        np.r_[O[1], O[1] - rcs*rca*nin[1]],
        np.r_[O[2], O[2] - rcs*rca*nin[2]],
        marker='o',
        ls='-',
        c='b',
    )
    ax.plot(
        np.r_[A[0], B[0]],
        np.r_[A[1], B[1]],
        np.r_[A[2], B[2]],
        marker='s',
        ls='-',
        c='k',
    )
    ax.plot(
        np.r_[Ex, Dxi],
        np.r_[Ey, Dyi],
        np.r_[Ez, Dzi],
        marker='x',
        ls='-',
        c='r',
    )
    print()
    return



def _debug_cylindrical(
    nin=None,
    erot=None,
    eax=None,
    nix=None,
    niy=None,
    niz=None,
    xxi=None,
    thetamax=None,
    xmax=None,
    roots=None,
    kk=None,
    eq=None,
    pt_x=None,
    pt_y=None,
    pt_z=None,
    pts_x=None,
    pts_y=None,
    pts_z=None,
    O=None,
    Dxi=None,
    Dyi=None,
    Dzi=None,
    rc=None,
    ii=None,
    ra=None,
    rb=None,
    **kwdargs,
):

    thetai = np.arctan2(
        (nin[1]*niz - nin[2]*niy)*eax[0]
        + (nin[2]*nix - nin[0]*niz)*eax[1]
        + (nin[0]*niy - nin[1]*nix)*eax[2],
        nix*nin[0] + niy*nin[1] + niz*nin[2],
    )

    # ----------------
    #  plot xx, theta

    fig = plt.figure(figsize=(15, 8))
    fig.suptitle(f"ii = {ii}")

    ax0 = fig.add_subplot(131)
    ax0.plot(
        kk, xxi, '.-b',
        kk, thetai, '.-r',
    )
    ax0.axhspan(-thetamax, thetamax, fc='r', ls='--', alpha=0.2)
    ax0.axhspan(-xmax, xmax, fc='b', ls='--', alpha=0.2)
    for rr in roots:
        ax0.axvline(rr, c='r', ls='--')

    # ----------------
    #  plot equation

    ax1 = fig.add_subplot(132)
    ax1.plot(
        kk, eq, '.-k',
        kk, ra*kk + rb, '-b',
    )
    ax1.axhline(0, ls='--', c='k')
    for rr in roots:
        ax1.axvline(rr, c='r', ls='--')

    # ----------------
    #  plot geometry

    ax2 = fig.add_subplot(133, projection='3d')
    ax2.plot(
        np.r_[pt_x, pts_x[ii]],
        np.r_[pt_y, pts_y[ii]],
        np.r_[pt_z, pts_z[ii]],
        '.-k',
    )

    # cents
    ax2.plot(
        np.r_[O[0], O[0] - rc*nin[0]],
        np.r_[O[1], O[1] - rc*nin[1]],
        np.r_[O[2], O[2] - rc*nin[2]],
        'ob',
    )

    # axe
    ax2.plot(
        O[0] + 5*xmax*np.r_[-1, 1] * eax[0],
        O[1] + 5*xmax*np.r_[-1, 1] * eax[1],
        O[2] + 5*xmax*np.r_[-1, 1] * eax[2],
        '--k',
    )

    # circles
    ang = thetamax*np.linspace(-1, 1, 11)
    ax2.plot(
        O[0] - xmax*eax[0] + rc*(-np.cos(ang)*nin[0] + np.sin(ang)*erot[0]),
        O[1] - xmax*eax[1] + rc*(-np.cos(ang)*nin[1] + np.sin(ang)*erot[1]),
        O[2] - xmax*eax[2] + rc*(-np.cos(ang)*nin[2] + np.sin(ang)*erot[2]),
        '--k',
    )
    ax2.plot(
        O[0] + xmax*eax[0] + rc*(-np.cos(ang)*nin[0] + np.sin(ang)*erot[0]),
        O[1] + xmax*eax[1] + rc*(-np.cos(ang)*nin[1] + np.sin(ang)*erot[1]),
        O[2] + xmax*eax[2] + rc*(-np.cos(ang)*nin[2] + np.sin(ang)*erot[2]),
        '--k',
    )

    Ex = pt_x + roots*(pts_x[ii] - pt_x)
    Ey = pt_y + roots*(pts_y[ii] - pt_y)
    Ez = pt_z + roots*(pts_z[ii] - pt_z)

    OEax = (Ex - O[0])*eax[0] + (Ey - O[1])*eax[1] + (Ez - O[2])*eax[2]
    Er = np.sqrt(
        ((Ex - O[0]) - OEax*eax[0])**2
        + ((Ey - O[1]) - OEax*eax[1])**2
        + ((Ez - O[2]) - OEax*eax[2])**2
    )
    print(Er)
    import pdb; pdb.set_trace()     # DB


def _debug_spherical(
    pt_x=None,
    pt_y=None,
    pt_z=None,
    pts_x=None,
    pts_y=None,
    pts_z=None,
    rr=None,
    O=None,
    rc=None,
    ABx=None,
    ABy=None,
    ABz=None,
    nin=None,
    e0=None,
    e1=None,
    **kwdargs,
):

    # get E
    Ex = pt_x + rr * ABx
    Ey = pt_y + rr * ABy
    Ez = pt_z + rr * ABz

    # get nout(E)
    nout_x = Ex - O[0]
    nout_y = Ey - O[1]
    nout_z = Ez - O[2]
    nout_norm = np.sqrt(nout_x**2 + nout_y**2 + nout_z**2)
    nout_x = nout_x / nout_norm
    nout_y = nout_y / nout_norm
    nout_z = nout_z / nout_norm

    # get D
    Dx = O[0] + rc*nout_x
    Dy = O[1] + rc*nout_y
    Dz = O[2] + rc*nout_z

    import matplotlib.pyplot as plt
    fig = plt.figure()
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection='3d')

    ax.plot(
        np.r_[pt_x, pts_x],
        np.r_[pt_y, pts_y],
        np.r_[pt_z, pts_z],
        '.-k',
    )

    for ii in range(len(rr)):
        l0, = ax.plot(
            np.r_[pt_x, Dx[ii], pts_x],
            np.r_[pt_y, Dy[ii], pts_y],
            np.r_[pt_z, Dz[ii], pts_z],
            '.-',
        )
        ax.plot(
            np.r_[Ex[ii], Dx[ii]],
            np.r_[Ey[ii], Dy[ii]],
            np.r_[Ez[ii], Dz[ii]],
            ls='--',
            marker='.',
            c=l0.get_color(),
        )

    # circles
    ax.plot(
        np.r_[O[0], O[0] - rc*nin[0]],
        np.r_[O[1], O[1] - rc*nin[1]],
        np.r_[O[2], O[2] - rc*nin[2]],
        'ob',
    )

    # circles
    theta = np.pi*np.linspace(-1, 1, 100)[1:]
    ax.plot(
        O[0] + rc*(np.cos(theta)*nin[0] + np.sin(theta)*e0[0]),
        O[1] + rc*(np.cos(theta)*nin[1] + np.sin(theta)*e0[1]),
        O[2] + rc*(np.cos(theta)*nin[2] + np.sin(theta)*e0[2]),
        '--k',
    )
    ax.plot(
        O[0] + rc*(np.cos(theta)*nin[0] + np.sin(theta)*e1[0]),
        O[1] + rc*(np.cos(theta)*nin[1] + np.sin(theta)*e1[1]),
        O[2] + rc*(np.cos(theta)*nin[2] + np.sin(theta)*e1[2]),
        '--k',
    )
    ax.plot(
        O[0] + rc*(np.cos(theta)*e0[0] + np.sin(theta)*e1[0]),
        O[1] + rc*(np.cos(theta)*e0[1] + np.sin(theta)*e1[1]),
        O[2] + rc*(np.cos(theta)*e0[2] + np.sin(theta)*e1[2]),
        '--k',
    )

    # check angles
    DA = np.r_[pt_x - Dx, pt_y - Dy, pt_z - Dz]
    DB = np.r_[pts_x - Dx, pts_y - Dy, pts_z - Dz]
    DE = np.r_[Ex - Dx, Ey - Dy, Ez - Dz]
    nA = np.linalg.norm(DA)
    nB = np.linalg.norm(DB)
    nE = np.linalg.norm(DE)
    print(np.sum(DA*DE) / (nA*nE))
    print(np.sum(DB*DE) / (nB*nE))

    import pdb; pdb.set_trace()     # DB



"""
Creating a geometry with Inkscape
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This is a tutorial that shows you how to draw a geometry in Inkscape
and load it in tofu.
"""

###############################################################################
# To see a tutorial on how to draw a vessel's geometry on Inkscape
# by tracing an image from a PDF file for example, and save it to a `svg` file,
# to load it and use it in tofu, you can check the video below. Basically, you
# just need to use straight *Bezier curves* to draw the closed polygons that
# will define the vessel and optionally the structures. To define a PFC
# structure, just add a fill color.
# You can also draw a line to automatically scale the configuration to a
# known measure.

###############################################################################

###############################################################################
# .. raw:: html
#
#    <div class="text-center">
#    <iframe width="560" height="315"
#    src="https://www.youtube.com/embed/MFwMZL7JjhI"
#    title="YouTube video player"
#    frameborder="0" allow="accelerometer; autoplay; clipboard-write;
#    encrypted-media; gyroscope; picture-in-picture" allowfullscreen>
#    </iframe>
#    </div>

###############################################################################
# Now, on how to load it and use it in tofu (also shown in the video). We start
# by importing `tofu`.

import tofu as tf

###############################################################################
# `tofu` provides a geometry helper function that allows creating a
# configuration from a `svg` file. Supposing you saved it in a file named
# `"myfirstgeom.svg"`, do
config = tf.geom.Config.from_svg("myfirstgeom.svg", Name="Test", Exp="Test")

###############################################################################
# It already gives you some information on what you loaded, but to see what it
# actually contains:
print(config)

###############################################################################
# To plot it, simply do
config.plot()

###############################################################################
# We can see that the z- and r-axis might not be exactly what we wanted. We
# might also want to adjust the scale. Tofu let's fix those parameters
# when loading the configuration.
config = tf.geom.Config.from_svg("myfirstgeom.svg", Name="Test", Exp="Test",
                                 z0=-140,
                                 r0=10,
                                 scale=0.5
                                 )
config.plot()

###############################################################################
# Or, even better, if you have a figure where there is known measure you can
# let tofu scale the figure automatically for you
config = tf.geom.Config.from_svg(
    "from_pdf.svg",
    Name="Traced from pdf",
    Exp="Test",
    res=10,
    point_ref1=(0.7, -2),
    point_ref2=(2.8, 2)
)
config.plot()
# sphinx_gallery_thumbnail_number = 3



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#

from libc.stdint cimport int64_t
cimport numpy as cnp
# ==============================================================================
# =  LINEAR MESHING
# ==============================================================================
cdef int64_t discretize_line1d_core(double* LMinMax, double dstep,
                                 double[2] DL, bint Lim,
                                 int mode, double margin,
                                 double** ldiscret_arr,
                                 double[1] resolution,
                                 int64_t** lindex_arr, int64_t[1] N) nogil

cdef void first_discretize_line1d_core(double* LMinMax,
                                       double dstep,
                                       double[1] resolution,
                                       int64_t[1] num_cells,
                                       int64_t[1] Nind,
                                       int[1] nL0,
                                       double[2] DL,
                                       bint Lim,
                                       int mode,
                                       double margin) nogil

cdef void second_discretize_line1d_core(double* LMinMax,
                                        double* ldiscret,
                                        int64_t* lindex,
                                        int nL0,
                                        double resolution,
                                        int64_t Nind) nogil

cdef void simple_discretize_line1d(double[2] LMinMax, double dstep,
                                   int mode, double margin,
                                   double** ldiscret_arr,
                                   double[1] resolution,
                                   int64_t[1] N) nogil

cdef void cythonize_subdomain_dl(DL, double[2] dl_array) # uses gil


# ==============================================================================
# =  Vessel's poloidal cut discretization
# ==============================================================================
cdef void discretize_vpoly_core(double[:, ::1] VPoly, double dstep,
                                int mode, double margin, double DIn,
                                double[:, ::1] VIn,
                                double** XCross, double** YCross,
                                double** reso, int64_t** ind,
                                int64_t** numcells, double** Rref,
                                double** XPolybis, double** YPolybis,
                                int[1] tot_sz_vb, int[1] tot_sz_ot,
                                int NP) nogil

# ------------------------------------------------------------------------------
# - Simplified version of previous algo
# ------------------------------------------------------------------------------
cdef void simple_discretize_vpoly_core(double[:, ::1] VPoly,
                                       int num_pts,
                                       double dstep,
                                       double** XCross,
                                       double** YCross,
                                       int[1] new_nb_pts,
                                       int mode,
                                       double margin) nogil

# ==============================================================================
# == LOS sampling
# ==============================================================================

# -- LOS sampling for a single ray ---------------------------------------------
cpdef int get_nb_imode(str imode)

cpdef int get_nb_dmode(str dmode)

cdef int los_get_sample_single(double los_kmin, double los_kmax,
                               double resol, int imethod, int imode,
                               double[1] eff_res, double** coeffs) nogil

# -- Calc signal utility function ---------------------------------------------
cdef call_get_sample_single_ani(double los_kmin, double los_kmax,
                                double resol,
                                int n_dmode, int n_imode,
                                double[1] eff_res,
                                int64_t[1] nb_rows,
                                double[:, ::1] ray_orig,
                                double[:, ::1] ray_vdir)

cdef cnp.ndarray[double,
                 ndim=2,
                 mode='c'] call_get_sample_single(double los_kmin,
                                                  double los_kmax,
                                                  double resol,
                                                  int n_dmode,
                                                  int n_imode,
                                                  double[1] eff_res,
                                                  int64_t[1] nb_rows,
                                                  double[:, ::1] ray_orig,
                                                  double[:, ::1] ray_vdir)

cdef int los_get_sample_core_const_res(int nlos,
                                       double* los_lim_min,
                                       double* los_lim_max,
                                       int n_dmode, int n_imode,
                                       double val_resol,
                                       double** coeff_ptr,
                                       double* dLr,
                                       int64_t* los_ind,
                                       int num_threads) nogil

cdef void los_get_sample_core_var_res(int nlos,
                                     double* los_lim_min,
                                     double* los_lim_max,
                                     int n_dmode, int n_imode,
                                     double* resol,
                                     double** coeff_ptr,
                                     double* dLr,
                                     int64_t* los_ind,
                                     int num_threads) nogil

cdef void los_get_sample_pts(int nlos,
                             double* ptx,
                             double* pty,
                             double* ptz,
                             double* usx,
                             double* usy,
                             double* usz,
                             double[:,::1] ray_orig,
                             double[:,::1] ray_vdir,
                             double* coeff_ptr,
                             int64_t* los_ind,
                             int num_threads) nogil

# ==============================================================================
# == Vmesh sampling
# ==============================================================================

# -- Vmesh utility functions --------------------------------------------------
cdef int  vmesh_disc_phi(int sz_r, int sz_z,
                         int64_t* ncells_rphi,
                         double phistep,
                         int ncells_rphi0,
                         double* disc_r,
                         double* disc_r0,
                         double* step_rphi,
                         double[::1] reso_phi_mv,
                         int64_t* tot_nc_plane,
                         int ind_loc_r0,
                         int ncells_r0,
                         int ncells_z,
                         int* max_sz_phi,
                         double min_phi,
                         double max_phi,
                         int64_t* sz_phi,
                         int64_t[:,::1] indi_mv,
                         double margin,
                         int num_threads) nogil

cdef int vmesh_get_index_arrays(int64_t[:, :, ::1] lnp,
                                int64_t[:, ::1] is_in_vignette,
                                int sz_r,
                                int sz_z,
                                int64_t* sz_phi) nogil

cdef void vmesh_assemble_arrays(int64_t[::1] first_ind_mv,
                                int64_t[:, ::1] indi_mv,
                                int64_t[:, ::1] is_in_vignette,
                                bint is_cart,
                                int sz_r,
                                int sz_z,
                                int64_t* lindex_z,
                                int64_t* ncells_rphi,
                                int64_t* tot_nc_plane,
                                double reso_r_z,
                                double* step_rphi,
                                double* disc_r,
                                double* disc_z,
                                int64_t[:,:,::1] lnp,
                                int64_t* phin,
                                double[::1] dv_mv,
                                double[::1] r_on_phi_mv,
                                double[:, ::1] pts_mv,
                                int64_t[::1] ind_mv,
                                int num_threads) nogil

cdef void vmesh_ind_init_tabs(int* ncells_rphi,
                              double* disc_r,
                              int sz_r, int sz_z,
                              double twopi_over_dphi,
                              double[::1] dRPhirRef,
                              int64_t* tot_nc_plane,
                              double** phi_mv,
                              int num_threads) nogil

cdef void vmesh_ind_cart_loop(int np,
                              int sz_r,
                              int64_t[::1] ind,
                              int64_t* tot_nc_plane,
                              int* ncells_rphi,
                              double* phi_tab,
                              double* disc_r,
                              double* disc_z,
                              double[:,::1] pts,
                              double[::1] res3d,
                              double reso_r_z,
                              double[::1] dRPhirRef,
                              int[::1] Ru,
                              double[::1] dRPhir,
                              int num_threads) nogil

cdef void vmesh_ind_polr_loop(int np,
                              int sz_r,
                              int64_t[::1] ind,
                              int64_t* tot_nc_plane,
                              int* ncells_rphi,
                              double* phi_tab,
                              double* disc_r,
                              double* disc_z,
                              double[:,::1] pts,
                              double[::1] res3d,
                              double reso_r_z,
                              double[::1] dRPhirRef,
                              int[::1] Ru,
                              double[::1] dRPhir,
                              int num_threads) nogil


# ==============================================================================
# == Solid Angles
# ==============================================================================
cdef int  sa_disc_phi(int sz_r, int sz_z,
                      int64_t* ncells_rphi,
                      double phistep,
                      double* disc_r,
                      double* disc_r0,
                      double* step_rphi,
                      int ind_loc_r0,
                      int ncells_r0,
                      int ncells_z,
                      int* max_sz_phi,
                      double min_phi,
                      double max_phi,
                      int64_t* sz_phi,
                      int64_t[:, ::1] indi_mv,
                      double margin,
                      int num_threads) nogil


cdef int sa_get_index_arrays(int64_t[:, ::1] lnp,
                             int64_t[:, ::1] is_in_vignette,
                             int sz_r,
                             int sz_z) nogil

cdef void sa_assemble_arrays(int block,
                             int use_approx,
                             double[:, ::1] part_coords,
                             double[::1] part_rad,
                             int64_t[:, ::1] is_in_vignette,
                             double[:, ::1] sa_map,
                             double[:, ::1] ves_poly,
                             double[:, ::1] ves_norm,
                             double[::1] ves_lims,
                             # int64_t[::1] lstruct_nlim,
                             int64_t[::1] lstruct_nlim,
                             double[::1] lstruct_polyx,
                             double[::1] lstruct_polyy,
                             double[::1] lstruct_lims,
                             double[::1] lstruct_normx,
                             double[::1] lstruct_normy,
                             # int64_t[::1] lnvert,
                             int64_t[::1] lnvert,
                             int nstruct_tot,
                             int nstruct_lim,
                             double rmin,
                             double eps_uz, double eps_a,
                             double eps_vz, double eps_b,
                             double eps_plane,
                             bint forbid,
                             int64_t[::1] first_ind_mv,
                             int64_t[:, ::1] indi_mv,
                             int sz_p,
                             int sz_r, int sz_z,
                             int64_t* ncells_rphi,
                             double reso_r_z,
                             double* disc_r,
                             double* step_rphi,
                             double* disc_z,
                             int64_t[:, ::1] ind_rz2pol,
                             int64_t* sz_phi,
                             double[::1] reso_rdrdz,
                             double[:, ::1] pts_mv,
                             int64_t[::1] ind_mv,
                             int num_threads)


# ##################################################################################
# ##################################################################################
#               Solid angle of a polygon
# ##################################################################################


cdef double comp_sa_tri(
    double A_x,
    double A_y,
    double A_z,
    double B_x,
    double B_y,
    double B_z,
    double C_x,
    double C_y,
    double C_z,
    double pt_x,
    double pt_y,
    double pt_z,
) nogil




import numpy as np
import scipy.constants as scpct
import scipy.special as scpsp
import datastock as ds


from . import _utils


__all__ = [
    'get_critical_dreicer_electric_fields',
    'get_normalized_momentum_distribution',
    'get_growth_source_terms',
]


# ##############################################################
# ##############################################################
#                 DEFAULTS
# ##############################################################


# see:
# https://docs.plasmapy.org/en/stable/notebooks/formulary/coulomb.html
_SIGMAP = 1.
_LNG = 20.


# ##############################################################
# ##############################################################
#                 Critical and Dreicer electric fields
# ##############################################################


def get_critical_dreicer_electric_fields(
    ne_m3=None,
    kTe_eV=None,
    lnG=None,
):

    # -------------
    # check input
    # -------------

    ne_m3, kTe_eV, lnG = _check_critical_dreicer(
        ne_m3=ne_m3,
        kTe_eV=kTe_eV,
        lnG=lnG,
    )

    # -------------
    # prepare
    # -------------

    # vacuum permittivity in C/(V.m), scalar
    eps0 = scpct.epsilon_0

    # custom computation intermediates C^2/(V^2.m^2), scalar
    pie02 = np.pi * eps0**2

    # electron charge (C), scalar
    e = scpct.e

    # electron rest energy (J = C.V), scalar
    mec2_CV = scpct.m_e * scpct.c**2

    # -------------
    # compute
    # -------------

    # critical electric field (V/m)
    Ec_Vm = ne_m3 * e**3 * lnG / (4 * pie02 * mec2_CV)

    # Dreicer electric field
    if kTe_eV is not None:
        Ed_Vm = Ec_Vm * (mec2_CV / e) / kTe_eV
    else:
        Ed_Vm = None

    # -------------
    # format output
    # -------------

    dout = {
        'E_C': {
            'data': Ec_Vm,
            'units': 'V/m',
        },
    }

    if Ed_Vm is not None:
        dout['E_D'] = {
            'data': Ed_Vm,
            'units': 'V/m',
        }

    return dout


def _check_critical_dreicer(
    ne_m3=None,
    kTe_eV=None,
    lnG=None,
):

    # -----------------
    # preliminary: lnG
    # -----------------

    if lnG is None:
        lnG = _LNG

    # -----------------
    # broadcastable
    # -----------------

    dparams, shape = ds._generic_check._check_all_broadcastable(
        ne_m3=ne_m3,
        kTe_eV=kTe_eV,
        lnG=lnG,
    )

    return [dparams[kk] for kk in ['ne_m3', 'kTe_eV', 'lnG']]


# ##############################################################
# ##############################################################
#                 Normalized Momentum Distribution
# ##############################################################


def get_normalized_momentum_distribution(
    momentum_normalized=None,
    # parameters
    ne_m3=None,
    Zeff=None,
    electric_field_par_Vm=None,
    energy_kinetic_max_eV=None,
    # optional
    lnG=None,
    sigmap=None,
    # options
    return_intermediates=None,
):
    """ Return the normalized RE momentum distribution, interpolated at pp

    Depends on:
        - pp: normalized kinetic momentum (variable)
              Assumed to ba a flat np.ndarray of shape (npp,)

    Parameters:
        - ne:_m3 background electron density (1/m3)
        - Zeff: effectove charge
        - Epar_Vm: parallel electric field (V/m)
        - Emax_eV: maximum kinetic energy (eV)
    All assumed to be broadcastable against each other

    Return a distribution of shape = (npp,) + shape of parameters

    Distribution is analytically nornalized

    ref:
    [1] Pandya et al., Physica Scripta 93, no. 11 (November 1, 2018): 115601

    Here we assume a pitch angle of 0:
        - p_perp = 0
        - p_par = pp
    """

    # -------------
    # check inputs
    # -------------

    (
        pp, ne_m3, Zeff,
        Epar_Vm, Emax_eV,
        sigmap, lnG,
        shape,
        return_intermediates,
    ) = _check_dist(
        pp=momentum_normalized,
        ne_m3=ne_m3,
        Zeff=Zeff,
        Epar_Vm=electric_field_par_Vm,
        Emax_eV=energy_kinetic_max_eV,
        sigmap=sigmap,
        return_intermediates=return_intermediates,
    )

    # -----------
    # initialize
    # -----------

    re_dist = np.full(shape, np.nan)

    # -------------
    # prepare
    # -------------

    # get momentum max from total energy eV.s/m - shape
    pmax = _utils.convert_momentum_velocity_energy(
        energy_kinetic_eV=Emax_eV,
    )['momentum_normalized']['data']

    # Critical electric field - shape
    Ec_Vm = get_critical_dreicer_electric_fields(
        ne_m3=ne_m3,
        kTe_eV=None,
        lnG=lnG,
    )['E_C']['data']

    # -------------
    # Intermediates
    # -------------

    # normalized electric field, adim
    Etild = Epar_Vm / Ec_Vm

    # ---------------------------
    # intermediate check on Etild
    # ---------------------------

    iok = Etild > 1.
    if np.any(iok):

        Ehat = (Etild[iok] - 1) / (1 + Zeff[iok])

        # adim
        Cz = np.sqrt(3 * (Zeff[iok] + 5) / np.pi)

        # critical momentum, adim
        pc = 1. / np.sqrt(Etild[iok] - 1.)

        # Cs
        Cs = (
            Etild[iok]
            - (
                ((1 + Zeff[iok])/4)
                * (Etild[iok] - 2)
                * np.sqrt(Etild[iok] / (Etild[iok] - 1))
            )
        )

        # -------------------
        # kwdargs to func

        kwdargs = {
            'sigmap': sigmap[iok],
            'pp': pp[iok],
            'pmax': pmax[iok],
            'Etild': Etild[iok],
            'Zeff': Zeff[iok],
            'Ehat': Ehat,
            'Cz': Cz,
            'Cs': Cs,
            'lnG': lnG[iok],
        }

        # ------------------
        # Compute

        # avalanche-dominated
        ioki = Etild[iok] > 5.
        if np.any(ioki):
            kwdargsi = {k0: v0[ioki] for k0, v0 in kwdargs.items()}
            iok0 = np.copy(iok)
            iok0[iok0] = ioki
            re_dist[iok0] = _re_dist_avalanche(**kwdargsi)

        # Dreicer-dominated
        ioki = (2 < Cs) & (Cs < 1 + Etild[iok])
        if np.any(ioki):
            kwdargsi = {k0: v0[ioki] for k0, v0 in kwdargs.items()}
            iok0 = np.copy(iok)
            iok0[iok0] = ioki
            re_dist[iok0] = _re_dist_dreicer(**kwdargsi)

        # --------------------------------
        # Set to 0 below critical momentum

        iout = np.copy(iok)
        iout[iok] = pp[iok] < pc
        re_dist[iout] = np.nan

    # -----------------------
    # no valid electric field

    else:
        Ehat = None
        Cz = None
        pc = None
        Cs = None

    # -------------
    # format output
    # -------------

    dout = {
        'dist': {
            'data': re_dist,
            'units': None,
        },
    }

    # ----------------------
    # optional intermediates

    if return_intermediates is True:
        dout.update({
            'Cs': {
                'data': Cs,
                'units': None,
            },
            'Cz': {
                'data': Cz,
                'units': None,
            },
            'Ec': {
                'data': Ec_Vm,
                'units': 'V/m',
            },
            'Etild': {
                'data': Etild,
                'units': None,
            },
            'Ehat': {
                'data': Ehat,
                'units': None,
            },
            'pc': {
                'data': pc,
                'units': None,
            },
        })

    return dout


def _check_dist(
    pp=None,
    ne_m3=None,
    Zeff=None,
    Epar_Vm=None,
    Emax_eV=None,
    sigmap=None,
    lnG=None,
    # options
    return_intermediates=None,
):

    # -----------------------
    # sigmap
    # -----------------------

    # Fermi decay width, dimensionless
    # [1] Pandya et al., Physica Scripta 93, no. 11 (November 1, 2018): 115601

    if sigmap is None:
        sigmap = _SIGMAP

    # -----------------
    # preliminary: lnG
    # -----------------

    if lnG is None:
        lnG = _LNG

    # -----------------------
    # all broadcastable
    # -----------------------

    dparams, shape = ds._generic_check._check_all_broadcastable(
        return_full_arrays=True,
        **locals(),
    )
    lk = ['pp', 'ne_m3', 'Zeff', 'Epar_Vm', 'Emax_eV', 'sigmap', 'lnG']
    lout = [dparams[k0] for k0 in lk]

    # ---------------------
    # return_intermediates
    # ---------------------

    return_intermediates = ds._generic_check._check_var(
        return_intermediates, 'return_intermediates',
        types=bool,
        default=False,
    )

    return lout + [shape, return_intermediates]


def _re_dist_avalanche(
    sigmap=None,
    pp=None,
    pmax=None,
    Ehat=None,
    Cz=None,
    lnG=None,
    # unused
    **kwdargs,
):

    # fermi decay factor, adim
    fermi = 1. / (np.exp((pp - pmax) / sigmap) + 1.)

    # distribution, adim
    re_dist = (
        (Ehat / (2*np.pi*Cz*lnG))
        * (1/pp)
        * np.exp(- pp / (Cz * lnG))
        * fermi
    )

    return re_dist


def _re_dist_dreicer(
    pp=None,
    Etild=None,
    Zeff=None,
    Cs=None,
    # unused
    **kwdargs,
):
    """ Distribution when primary RE generation is dominant
    see eq (7) in:
        Pandya et al. 2018

    """

    # assumption
    p_perp = 0.
    p_par = pp

    # pper2par
    pperp2par = p_perp**2 / p_par

    # Hypergeometric confluent Kummer function
    term1 = 1 - Cs / (Etild + 1)
    term2 = ((Etild + 1) / (2.*(1. + Zeff))) * pperp2par
    F1 = scpsp.hyp1f1(term1, 1, term2)

    # ppar_exp_inv
    ppar_exp_inv = 1./(p_par**((Cs - 2.) / (Etild - 1.)))

    # exponential
    exponential = np.exp(-((Etild + 1) / (2 * (1 + Zeff))) * pperp2par)

    # distribution
    re_dist = ppar_exp_inv * exponential * F1

    return re_dist


# ##############################################################
# ##############################################################
#            Primary & secondary growth source terms
# ##############################################################


def get_growth_source_terms(
    ne_m3=None,
    lnG=None,
    Epar_Vm=None,
    kTe_eV=None,
    Zeff=None,
):
    """ Return the source terms in the RE dynamic equation

    S_primary: dreicer growth  (1/m3/s)

    S_secondary: avalanche growth (1/s)

    """

    # -------------
    # check inputs
    # -------------

    ne_m3, lnG, Epar_Vm, kTe_eV, Zeff = _check_growth(
        ne_m3=ne_m3,
        lnG=lnG,
        Epar_Vm=Epar_Vm,
        kTe_eV=kTe_eV,
        Zeff=Zeff,
    )

    # -------------
    # prepare
    # -------------

    # vacuum permittivity in C/(V.m), scalar
    eps0 = scpct.epsilon_0

    # charge C
    e = scpct.e

    # mec2 (J = CV)
    mec2_CV = scpct.m_e * scpct.c**2

    # mec C.V.s/m
    mec = mec2_CV / scpct.c

    # me2c3 J**2 / (m/s) = C^2 V^2 s / m
    me2c3 = mec2_CV**2 / scpct.c

    # Dreicer electric field - shape
    dEcEd = get_critical_dreicer_electric_fields(
        ne_m3=ne_m3,
        kTe_eV=kTe_eV,
        lnG=lnG,
    )

    Ec_Vm = dEcEd['E_C']['data']
    Ed_Vm = dEcEd['E_D']['data']

    # -------------
    # pre-compute
    # -------------

    # term1 (m^3/s)
    term1 = e**4 * lnG / (4 * np.pi * eps0**2 * me2c3)

    # term2 - unitless (convert kTe_eV => J)
    term2 = (mec2_CV / (2. * kTe_eV * e))**1.5

    # term3 - unitless
    term3 = (Ed_Vm / Epar_Vm)**(3*(1. + Zeff) / 16.)

    # exp - unitless
    exp = np.exp(
        -Ed_Vm / (4.*Epar_Vm) - np.sqrt((1. + Zeff) * Ed_Vm / Epar_Vm)
    )

    # sqrt - unitless
    sqrt = np.sqrt(np.pi / (3 * (5 + Zeff)))

    # -------------
    # Compute
    # -------------

    # 1/m^3/s
    S_primary = ne_m3**2 * term1 * term2 * term3 * exp

    # 1/s   (C / C.V.s/m * V.m)
    S_secondary = sqrt * (e / mec) * (Epar_Vm - Ec_Vm) / lnG

    # -------------
    # format output
    # -------------

    dout = {
        'S_primary': {
            'data': S_primary,
            'units': '1/m3/s',
        },
        'S_secondary': {
            'data': S_secondary,
            'units': '1/s',
        },
    }

    return dout


def _check_growth(
    ne_m3=None,
    lnG=None,
    Epar_Vm=None,
    kTe_eV=None,
    Zeff=None,
):

    # -----------------
    # preliminary: lnG
    # -----------------

    if lnG is None:
        lnG = _LNG

    # -----------------------
    # all broadcastable
    # -----------------------

    dparams, shape = ds._generic_check._check_all_broadcastable(
        return_full_arrays=False,
        **locals(),
    )
    lk = ['ne_m3', 'lnG', 'Epar_Vm', 'kTe_eV', 'Zeff']
    lout = [dparams[k0] for k0 in lk]

    return lout
